
import unittest   # The test framework
from src.dphelper.helper import DPHelper


class TestDPHelperService(unittest.TestCase):
 
    def test_default_api(self):
        dphelper = DPHelper()
        self.assertEqual(dphelper.get_backend_url(), 'https://dphelper.dataplatform.lt')

    def test_api(self):
        dphelper = DPHelper(backend_url='https://dphelper.dataplatform.lt')
        greeting = dphelper.get_backend_greeting()
        self.assertEqual(greeting, 'backend')
        self.assertEqual(dphelper.get_backend_url(), 'https://dphelper.dataplatform.lt')

    def test_legacy_api(self):
        dphelper = DPHelper(backend_url='https://api.dataplatform.lt')
        greeting = dphelper.get_backend_greeting()
        self.assertEqual(greeting, 'backend')
        self.assertEqual(dphelper.get_backend_url(), 'https://api.dataplatform.lt')

    def test_dp_id_old(self):
        dphelper = DPHelper(backend_url='https://api.dataplatform.lt')
        self.assertEqual(dphelper.get_dp_url(), 'https://api.dataplatform.lt')
      
    def test_dp_id_new(self):
        dphelper = DPHelper(backend_url='https://dphelper.dataplatform.lt')
        self.assertEqual(dphelper.get_dp_url(), 'https://api.dataplatform.lt')

if __name__ == '__main__':
    unittest.main()



import typing as _typing
import logging as _logging
from typing import Optional as _Optional, List as _List, Tuple as _Tuple, Union as _Union

import requests as _requests
import concurrent.futures
from requests import Response as _Response
from requests.exceptions import ConnectionError as _ConnectionError
from time import sleep as _sleep
from urllib import parse as _parse
from fake_headers import Headers as _Headers

from ..config import Config as __Config


class FetchModule(__Config):
    def create_headers(
        self, authority: _Optional[str] = None, extra_headers: dict = {}
    ) -> dict:
        """create prefilled headers with custom authority value, append extra_headers.
        :param authority: what authority to use? If None, then no authority will be set.
        :param extra_headers: dict of headers that shall be appended
        :return: ready to use headers
        """
        headers = {}
        if authority is not None:
            headers["authority"] = authority
        headers.update(self.DEFAULT_HEADERS)
        headers.update(extra_headers)
        return headers

    def _get_headers_or_generate(self, headers: _Optional[dict] = ...) -> dict:
        "return headers or generate new ones"
        if isinstance(headers, dict) or headers is None:
            return headers
        new_headers = _Headers().generate()
        # generates:
        # 'Accept': '*/*'
        # 'Connection': 'keep-alive'
        # 'User-Agent': random_agent
        new_headers.update(self.HEADERS_UPDATE_ON_EMPTY)
        return new_headers

    @staticmethod
    def _compare_headers(left_headers: dict, right_headers: dict) -> str:
        """returns headers differences for debug purposes"""
        normalize = lambda headers: (  # keys to lowercase
            None
            if headers is None
            else {str(key).lower(): value for key, value in headers.items()}
        )
        left_headers = normalize(left_headers)
        right_headers = normalize(right_headers)
        if left_headers == right_headers:
            return "Headers equal"
        elif left_headers is None:
            return "Left-header is None"
        elif right_headers is None:
            return "Right-header is None"

        # what is unique in left-headers? in right-headers? what is unequal?
        exclusives = []
        unequalities = []
        left_keys, right_keys = set(left_headers), set(right_headers)
        exclusives.extend(
            f"exclusive to left-headers: {key}" for key in left_keys - right_keys
        )
        exclusives.extend(
            f"exclusive to right-headers: {key}" for key in right_keys - left_keys
        )

        for key in left_keys.intersection(right_keys):  # shared keys
            if left_headers[key] != right_headers[key]:
                unequalities.append(f"left-headers != right-headers at ({key}):")
                unequalities.append(f"({left_headers[key]}) != ({right_headers[key]})")

        return "\n".join(unequalities + exclusives)

    def __get_proxy_full_url(self):
        url = _parse.urljoin(self.BACKEND_URL, self.BACKEND_UTILS_PATH)
        return _parse.urljoin(url, "proxy/")

    def _request_or_repeat(self, function: callable, *args, **kwargs) -> _Response:
        """Makes a request (call provided function with args and kwargs).
        * If it fails because of ConnectionError, retries;
        * If it fails because status_code >= 500, retries by default;
        retry count and sleep times are specified in helper.py, config.py

        :param function: function to call; expected to get response object from it.
        :param *args: what args to pass to function.
        :param **kwargs: what kwargs to pass to function.
        :return: `Response` or raises RuntimeError after all failures
        """

        failures_count = 0
        sleep_time = self.REQUEST_INITIAL_SLEEP_AFTER_FAILURE_IN_S
        failure_reason = ""
        response = None

        def _handle_failure():
            nonlocal failures_count, sleep_time
            sleep_time = (
                sleep_time**self.REQUEST_SLEEP_INCREASE_POWER
                if failures_count > 0
                else sleep_time
            )
            failures_count = failures_count + 1

        while failures_count < self.REQUEST_DROP_ON_FAILURE_COUNT:
            if failures_count > 0:
                if self.IS_VERBOSE:
                    _logging.error(
                        f"Request failed: {failure_reason}. Retrying in {sleep_time} seconds"
                    )
                _sleep(sleep_time)
            try:
                response: _Response = function(*args, **kwargs)
                is_blocked = (self.IS_CRASH_ON_BLOCKED or self.IS_RETRY_ON_BLOCKED) and self._get_is_request_blocked(response) 
                if is_blocked and self.IS_CRASH_ON_BLOCKED:
                    raise Exception(self._make_blocked_error_text(response) + response.text)
                elif is_blocked and self.IS_RETRY_ON_BLOCKED:
                    raise ValueError
                for range in self.RETRYABLE_STATUS_CODE_RANGES:
                    if range[0] <= response.status_code <= range[1]:
                        raise RuntimeError
                return response
            except _ConnectionError:
                _handle_failure()
                failure_reason = "Server inexistant or check Network settings"
            except RuntimeError:
                _handle_failure()
                failure_reason = f"Remote server had an {'internal' if response is None else f'http.{response.status_code}'} error"
            except ValueError:
                _handle_failure()
                failure_reason = f"Remote server blocked our request with status http.{'???' if response is None else response.status_code}. Text: {'???' if response is None else response.text}"

        raise RuntimeError(
            f"[internal] Could not make connection for {self.REQUEST_DROP_ON_FAILURE_COUNT} times. Last reason of failure: {failure_reason}"
        )

    @staticmethod
    def _response_ok_or_raise(response: _requests.Response) -> None:
        """Raises ValueError if response returned failure code"""
        try:
            response.raise_for_status()
        except Exception as ex:
            raise ValueError(
                "The response was not ok. "
                f"Response code: {response.status_code} "
                f"response text: {response.text}"
            ) from ex

    def fetch_html(
        self,
        url: str,
        *,
        proxy_provider: _Optional[str] = ...,
        params: _Optional[dict] = None,
        headers: _Optional[dict] = ...,
        data: _typing.Any = None,
        json: _typing.Any = None,
        cookies: _Optional[dict] = None,
        method: str = "GET",
        is_repeat_on_failure: bool = True,
        proxy_params: _Optional[dict] = ...,
        stream: _Optional[bool] = None,
        verify: _Optional[bool] = None,
    ) -> _requests.Response:
        """returns Response object.
        :param proxy_provider: "scrapingbee" or "any" or your secret provider
        :param method: one of "POST" "GET" "PUT" "PATCH" "DELETE"
        """
        headers = self._get_headers_or_generate(headers)
        # no proxy_provider given? check is one given via DPHelper(proxy_provider="qqq")
        proxy_provider = (
            self.PROXY_PROVIDER if proxy_provider is ... else proxy_provider
        )
        proxy_params = self.PROXY_PARAMS if proxy_params is ... else proxy_params
        # to use default proxy provider, proxy_provider must be empty string or string "any".
        # using None would not use proxy, would make a direct request
        proxy_provider = "" if str(proxy_provider).lower() == "any" else proxy_provider

        # prepare perform_method and call it with request_kwargs
        if proxy_provider is None:
            # no proxy
            perform_method = (
                _requests.request  # request once
                if not is_repeat_on_failure
                else lambda *args, **kwargs: self._request_or_repeat(
                    _requests.request, *args, **kwargs
                )  # request and retry on 5xx, connection errors
            )

            request_kwargs = {
                "method": method,
                "url": url,
                "params": params,
                "headers": headers,
                "cookies": cookies,
                "data": data,
                "json": json,
                "stream": stream,
                "verify": verify,
            }
        else:
            # proxy yes
            # whoops cookies
            if cookies is not None:
                raise NotImplementedError(
                    "Current ",
                    self.PACKAGE_NAME,
                    " version does not support cookies in Proxy.",
                )
            # verify the proxy?
            if verify and self.IS_VERBOSE:
                _logging.warning("`verify` argument is ignored when using proxy")
            # perform_method + request_kwargs
            perform_method = (
                _requests.post  # once
                if not is_repeat_on_failure
                else lambda *args, **kwargs: self._request_or_repeat(
                    _requests.post,
                    *args,
                    **kwargs,
                )
            )

            request_kwargs = {
                "url": self.__get_proxy_full_url(),
                "params": {"provider": str(proxy_provider)},
                "json": {
                    "url": url,
                    "headers": headers,
                    "params": params,
                    "data": data,
                    "json": json,
                    "method": method,
                    "proxy_params": proxy_params,
                },
                "stream": stream,
            }

        # do request
        response = perform_method(**request_kwargs)
        if self.IS_VERBOSE or self.IS_CRASH_ON_BLOCKED:
            is_blocked = self._get_is_request_blocked(response)
            error = self._make_blocked_error_text(response)
            if self.IS_CRASH_ON_BLOCKED and is_blocked:
                raise Exception(error + response.text)
            elif self.IS_VERBOSE and is_blocked:
                _logging.warning(error)
        return response

    def from_url(
        self,
        url,
        verify=True,
        headers: _Optional[dict] = ...,
        *,
        proxy_provider: _Optional[str] = ...,
        proxy_params: _Optional[dict] = ...,
        encoding: _Optional[str] = None,
        trim_whitespaces=True,
    ):
        "return content with stripped trailing whitespaces (newlines) from each line"
        response = self.fetch_html(
            url,
            verify=verify,
            headers=headers,
            proxy_provider=proxy_provider,
            proxy_params=proxy_params,
        )

        # ISO-8859-1 is default encoding for html < 5
        # encoding order: user supplied one > one from response > default utf-8
        encoding_by_response = (
            "utf-8" if response.encoding in [None, "ISO-8859-1"] else response.encoding
        )
        encoding = encoding or encoding_by_response

        if not trim_whitespaces:
            return response.content.decode(encoding)
        else:
            return "".join(
                [line.decode(encoding).strip() for line in response.iter_lines()]
            )

    def crawl(self, urls, parser, max_concurrent=50, param_func=None):
        results = {}

        # Define a function that fetches and parses an URL, to be run in a thread
        def fetch_and_parse(url: str) -> None:
            try:
                params = param_func(url) if param_func else {}
                html = self.fetch_html(url, **params).text
                results[url] = parser(html)
            except Exception:
                results[url] = None

        # Use a ThreadPoolExecutor to run fetch_and_parse in parallel for all URLs
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=max_concurrent
        ) as executor:
            executor.map(fetch_and_parse, urls)

        return results

    def upload_all_images(self, urls, max_concurrent=50):
        # Bulk upload of images
        results = {}

        def upload_image(url: str) -> bytes:
            try:
                results[url] = self.upload_image_from_url(url)
            except Exception:
                results[url] = None

        with concurrent.futures.ThreadPoolExecutor(
            max_workers=max_concurrent
        ) as executor:
             executor.map(upload_image, urls)
    
        return results
    
    @staticmethod
    def _make_status_code_ranges(status_codes: _List[_Union[int, str]]) -> _List[_Tuple[int, int]]:
        """Converts status codes to ranges. Example: [200, 5xx] -> [(200, 200), (500, 599)]"""
        status_codes = list(set(status_codes))
        ranges = []
        for code in status_codes:
            if isinstance(code, int):
                ranges.append((code, code))
            elif isinstance(code, str):
                code: str
                ranges.append(
                    (
                        int(code.lower().replace("x", "0")),
                        int(code.lower().replace("x", "9"))
                    )
                )
            else:
                raise ValueError(f"Invalid status code: {code}")
        return ranges
    
    def _get_is_request_blocked(self, response: _requests.Response) -> bool:
        """Rates how likely the response is blocked (0 - 100) and then returns boolean"""
        text = response.text.lower()
        rating = 0
        if response.status_code in [429, 403, 400] or response.status_code >= 500:
            rating += 30
        for keyword in self.BLOCKED_KEYWORDS:
            if keyword in text:
                rating += 20
        rating = min(rating, 100)
        return rating >= self.BLOCKED_TRESHOLD
    
    @staticmethod
    def _make_blocked_error_text(response: _requests.Response) -> str:
        """Returns error text for blocked request"""
        return f"Our Request to {response.request.url} was potentially blocked by remote server. Response status code: http {response.status_code} . "


import unittest   # The test framework
from src.dphelper.helper import DPHelper

class TestDPHelperImageService(unittest.TestCase):

    def test_single_image_upload(self):
        dphelper = DPHelper(api_key='ADD_YOUR_KEY_TO_TEST')
        # self.assertEqual(dphelper.get_key_for_image_service(), 'ADD_YOUR_API_KEY_TO_TEST')
        result = dphelper.upload_image_from_url('https://www.google.com/images/branding/googlelogo/2x/googlelogo_color_92x30dp.png')
        self.assertGreater(len(result['id']), 0)
   
    def test_bulk_image_upload(self):
        dphelper = DPHelper(api_key='ADD_YOUR_KEY_TO_TEST')
        image_urls = [
            'https://www.google.com/images/branding/googlelogo/2x/googlelogo_color_92x30dp.png',
            'https://www.google.com/images/branding/googlelogo/1x/googlelogo_color_92x30dp.png',
        ]
        results = dphelper.upload_all_images(image_urls, max_concurrent=10)
        self.assertGreater(len(results["https://www.google.com/images/branding/googlelogo/2x/googlelogo_color_92x30dp.png"]["id"]), 5)

if __name__ == '__main__':
    unittest.main()




import unittest   # The test framework
from src.dphelper.helper import DPHelper

class TestDPHelperSnapshots(unittest.TestCase):
   
    def test_test_basic(self):
        dphelper = DPHelper()
        response_full = dphelper.snapshot.get_latest(by_challenge_id=140) 
        data_result = response_full.result
        self.assertGreater(len(data_result), 0)
  
if __name__ == '__main__':
    unittest.main()



import json
import os
import logging as _logging
from requests.structures import CaseInsensitiveDict


class Config:
    def __init__(self) -> None:
        # defining init, so IDE will think that all
        # Config instances, sub-instances has such inner values set
        self._init_config_values()

    def _init_config_values(self) -> None:
        self.PACKAGE_NAME = "dphelper"
        "Update this name after changing name in pyproject.toml"

        self.REQUEST_DROP_ON_FAILURE_COUNT = 4
        "How many `didnt connect` / `status >= 500` should occur sequentially before giving-up-on-request"

        self.REQUEST_INITIAL_SLEEP_AFTER_FAILURE_IN_S = 3
        "How many seconds to sleep after first request failure"

        self.REQUEST_SLEEP_INCREASE_POWER = 2
        "sleep time will be increased by this power. time ^ power"

        self.DP_URL = "https://dphelper.dataplatform.lt/"
        "url of server that stores snapshots, challenges, code runs. do not include ending slash"

        self.BACKEND_URL = "https://api.dataplatform.lt"
        "url of server that stores snapshots, challenges, code runs. do not include ending slash"

        self.BACKEND_SNAPSHOT_PATH = f"snapshots/"
        "backend url root for snapshots. it might be prepended in actual usage"

        self.BACKEND_UTILS_PATH = f"v1/utils/"
        "backend url root for utils. it might be appended in actual usage"

        self.IS_VERBOSE = False
        "verbose errors, warnings, etc.?"

        self.DEFAULT_HEADERS = {
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "accept-language": "en-GB,en-US;q=0.9,en;q=0.8,lt;q=0.7",
            "cache-control": "no-cache",
            "pragma": "no-cache",
            "sec-ch-ua": '"Not_A Brand";v="99", "Google Chrome";v="109", "Chromium";v="109"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"macOS"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "none",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/109.0.0.0 Safari/537.36",
        }

        self.HEADERS_UPDATE_ON_EMPTY = {
            "Referer": "https://google.com",
            "Cache-Control": "max-age=0",
            "Upgrade-Insecure-Requests": "1",
        }
        "what headers to append when generating auto-headers? auto-headers already include user agent, accept, connection"

        self.PROXY_PROVIDER = None
        "what proxy provider to use for outter requests? Snapshots would still be fetched directly"

        self.PROXY_PARAMS = {}
        "what params to use for proxying"

        self.RETRYABLE_STATUS_CODE_RANGES = [(429, 429), (500, 599)]
        "what status code ranges are retryable?"

        self.IS_RETRY_ON_BLOCKED = False
        "retry on blocked requests? Something like 'Checking your browser'"

        self.IS_CRASH_ON_BLOCKED = False
        "raise exception on blocked request? Something like 'Checking your browser'"

        self.BLOCKED_KEYWORDS = [
            x.lower()
            for x in [
                "(encodeURI('challenge='",
                ",jsChallengeUrl)",
                "/jschallenge",
                "Checking your browser",
                "Just a moment",
                "blocked",
            ]
        ]
        "what keywords to look for in response to detect blocked request"

        self.BLOCKED_TRESHOLD = 50
        "rating varies from 0 to 100. From what value and above to consider request blocked?"

        self.SECRETS_FILE_PATH = "secrets.dphelper.json"
        'where to look for secrets file. Format: json. Example content: {"something": "value", "smth2": "value2"}'

        self.SECRET = dict()
        "dict with secrets. It is populated from secrets file"

    def set_backend_url(self, new_backend_url: str) -> None:
        """Sets backend url. Do not forget http, https thingies.

        :param new_backend_url: the url where backend is located
        """
        if new_backend_url is None or new_backend_url == "":
            raise ValueError("Cannot set backend url to None or empty string")

        self.BACKEND_URL = new_backend_url

    def set_dp_url(self, new_dp_url: str) -> None:
        """Sets dp url. Do not forget http, https thingies.

        :param new_backend_url: the url where backend is located
        """
        if new_dp_url is None or new_dp_url == "":
            raise ValueError("Cannot set dp url to None or empty string")

        self.DP_URL = new_dp_url

    def populate_secrets_from_file(
        self, *, is_environ_prioritized=False, is_load_environ=True
    ) -> None:
        """Populates secrets dict from file. File should be in json format. Example content: {"something": "value", "smth2": "value2"}

        :param is_environ_prioritized: on True order: loads secrets, loads environ (overwrites some secrets)
        """

        try:
            environ_dict = CaseInsensitiveDict(os.environ if is_load_environ else {})
            if not os.path.exists(self.SECRETS_FILE_PATH):
                if self.IS_VERBOSE:
                    _logging.warning(
                        f"Secrets file not found at '{os.path.abspath(os.path.join(os.getcwd(), self.SECRETS_FILE_PATH))}'.{' Thus DPHelper will initialize SECRET from <os.environ> only.' if is_load_environ else ''}"
                    )
                self.SECRET = environ_dict
                return
            with open(self.SECRETS_FILE_PATH, "r") as f:
                secret_dict = {
                    str(key): str(value) for key, value in json.load(f).items()
                }
                main_dict, updater_dict = (
                    (secret_dict, environ_dict)
                    if is_environ_prioritized
                    else (environ_dict, secret_dict)
                )
                self.SECRET = CaseInsensitiveDict(main_dict)
                self.SECRET.update(updater_dict)
        except Exception as e:
            raise Exception(
                f"Failed to load secrets from file {self.SECRETS_FILE_PATH}. Error: {str(e)}"
            )

    def export_into_environ(self):
        for k, v in self.SECRET.items():
            os.environ[str(k)] = str(v)



import typing as _typing
from typing import Optional as _Optional, List as _List
import urllib.parse as _parse

import requests as _requests

from .. import schemas as _schemas
from ..config import Config as _Config
from ..connection.fetch import FetchModule as _FetchModule
import datetime


class SnapshotModule:
    def __init__(
        self, config: _Config, fetcher: _FetchModule, api_key: str = None
    ) -> None:
        """
        :param config: from here config values will be read: `config.DP_URL`
        :param fetcher: will use this object for fetching: `fetcher._request_or_repeat(...)`
        """
        self.config = config
        self.fetcher = fetcher
        self.API_KEY = api_key

    def __get_snapshot_url(self, path: str):
        """path will be joined with dp url"""
        # appending "/" to root so path gets joined, not overwriten
        # >>> "http://q.w/r" == _parse.urljoin("http://q.w/e", "r")
        # >>> "http://q.w/e/r" == _parse.urljoin("http://q.w/e" + "///////", "r")
        url = _parse.urljoin(
            self.config.DP_URL + "/", self.config.BACKEND_SNAPSHOT_PATH
        )
        api_url = _parse.urljoin(url + "/", path)

        if self.API_KEY is not None:
            params = {"api_key": self.API_KEY}
            api_url = api_url + "?" + _parse.urlencode(params)

        return api_url

    def get_meta_by_id(self, snapshot_id: int) -> _schemas.SnapshotMeta:
        "gets meta info about snapshot (by snapshot id)"

        response = self.fetcher._request_or_repeat(
            _requests.get, url=self.__get_snapshot_url(f"{snapshot_id}/head")
        )
        self.fetcher._response_ok_or_raise(response)
        return _schemas.SnapshotMeta(**response.json())

    def __get_result_by_url_or_id(
        self,
        *,
        is_loadable: _Optional[bool] = None,
        result_file_url: _Optional[str] = None,
        snapshot_id: _Optional[int] = None,
    ) -> _typing.Any:
        "gets snapshot result by file url or id. if result is not loadable, {} will be returned"
        if result_file_url is None and snapshot_id is None:
            raise ValueError(
                "Cant get snapshot result without args. Please give file url of snapshot data or snapshot_id"
            )
        if is_loadable == False:
            # why bother downloading if it will be blank / corrupted / etc. ?
            # or should exception be raised?
            return {}
        is_preview = False
        if result_file_url is None:
            is_preview = True  # response will have to be unwrapped
            # maybe old snapshot, so ask for preview, expecting to get full snapshot result
            result_file_url = self.__get_snapshot_url(f"{snapshot_id}/preview")
        response = self.fetcher._request_or_repeat(_requests.get, url=result_file_url)
        self.fetcher._response_ok_or_raise(response)
        return (
            response.json() if not is_preview else response.json()["json_data_preview"]
        )

    def get_result_by_id(self, snapshot_id: int) -> _typing.Any:
        "gets snapshot result by snapshot id. result might be any type but usually it is a list of objects"
        # get meta
        meta = self.get_meta_by_id(snapshot_id)
        # now result
        return self.__get_result_by_url_or_id(
            is_loadable=meta.is_result_json_loadable,
            result_file_url=meta.result_file_url,
            snapshot_id=snapshot_id,
        )

    def get_by_id(self, snapshot_id: int) -> _schemas.Snapshot:
        "gets snapshot meta + snapshot result as a whole object. snapshot result is placed inside `result` attribute"
        meta = self.get_meta_by_id(snapshot_id)
        _dict = meta.model_dump() if hasattr(meta, "model_dump") else meta.dict()
        return _schemas.Snapshot(
            **_dict,
            result=self.__get_result_by_url_or_id(
                is_loadable=meta.is_result_json_loadable,
                result_file_url=meta.result_file_url,
                snapshot_id=snapshot_id,
            ),
        )

    def __raise_if_not_implemented_filter(
        self,
        *,
        by_is_verified: _Optional[bool] = None,
        by_validation_statuses: _Optional[_List[str]] = None,
        **_,
    ) -> None:
        "raises NotImplemented if one of the filters is not implemented"
        if by_is_verified is not None:
            raise NotImplementedError("Sorry, filter is_verified not implemented")
        elif by_validation_statuses is not None:
            raise NotImplementedError(
                "Sorry, filter validation_statuses not implemented"
            )

    def __args_to_filter_params(
        self,
        *,
        by_challenge_id: _Optional[int] = None,
        by_user_id: _Optional[int] = None,
        by_is_verified: _Optional[bool] = None,
        by_validation_statuses: _Optional[_List[str]] = None,
        by_is_from_robot: _Optional[bool] = None,
        **_,
    ) -> dict:
        "remaps kwargs so key names match backend names"
        return {
            "challenge_id": by_challenge_id,
            "user_id": by_user_id,
            "is_verified": by_is_verified,
            "validation_statuses": by_validation_statuses,
            "is_from_code_run": by_is_from_robot,
        }

    def get_latest_meta(
        self,
        by_challenge_id: _Optional[int] = None,
        *,
        by_user_id: _Optional[int] = None,
        by_is_verified: _Optional[bool] = None,
        by_validation_statuses: _Optional[_List[str]] = None,
        by_is_from_robot: _Optional[bool] = None,
        max_stale_days: _Optional[int] = None,
    ) -> _schemas.SnapshotMeta:
        """gets latest snapshot meta by specified filters.

        :param by_challenge_id: to what challenge shall snapshot belong? If None, any will match.
        :param by_user_id: to what user shall snapshot belong? If None, any will match.
        :param by_is_verified: shall the snapshot be moderator-approved? If None, any will match.
        :param by_validation_statuses: searched statuses of snapshot validation. If None, any will match.
        :param by_is_from_robot: shall the snapshot be generated from robot code run (True) or uploaded by human (False)? If None, any will match.
        """
        # check, is each filter supported / implemented.
        # at this point locals() will have dict of supplied args and their values
        # using __class__ so we can reuse self=smth from **locals()
        self.__class__.__raise_if_not_implemented_filter(**locals())
        response = self.fetcher._request_or_repeat(
            _requests.get,
            url=self.__get_snapshot_url("latest/head/"),
            params=self.__class__.__args_to_filter_params(**locals()),
        )
        self.fetcher._response_ok_or_raise(response)
        meta = _schemas.SnapshotMeta(**response.json())

        if max_stale_days is not None and meta.date_created is not None:
            if (datetime.datetime.now() - meta.date_created).days > max_stale_days:
                raise ValueError(
                    f"Snapshot is too old. It was created {meta.date_created}."
                )
        return meta

    def get_latest_result(
        self,
        by_challenge_id: _Optional[int] = None,
        *,
        by_user_id: _Optional[int] = None,
        by_is_verified: _Optional[bool] = None,
        by_validation_statuses: _Optional[_List[str]] = None,
        by_is_from_robot: _Optional[bool] = None,
        max_stale_days: _Optional[int] = None,
    ) -> _typing.Any:
        """gets latest snapshot result by specified filters

        :param by_challenge_id: to what challenge shall snapshot belong? If None, any will match.
        :param by_user_id: to what user shall snapshot belong? If None, any will match.
        :param by_is_verified: shall the snapshot be moderator-approved? If None, any will match.
        :param by_validation_statuses: searched statuses of snapshot validation. If None, any will match.
        :param by_is_from_robot: shall the snapshot be generated from robot code run (True) or uploaded by human (False)? If None, any will match.
        """
        # at this point locals() will have dict of supplied args and their values
        # using __class__ so we can reuse self=smth from **locals()
        meta = self.__class__.get_latest_meta(**locals())
        return self.__get_result_by_url_or_id(
            is_loadable=meta.is_result_json_loadable,
            result_file_url=meta.result_file_url,
            snapshot_id=meta.id,
        )

    def get_latest(
        self,
        by_challenge_id: _Optional[int] = None,
        *,
        by_user_id: _Optional[int] = None,
        by_is_verified: _Optional[bool] = None,
        by_validation_statuses: _Optional[_List[str]] = None,
        by_is_from_robot: _Optional[bool] = None,
        max_stale_days: _Optional[int] = None,
    ) -> _schemas.Snapshot:
        """gets latest snapshot meta+result by specified filters. result is stored inside `result` attribute

        :param by_challenge_id: to what challenge shall snapshot belong? If None, any will match.
        :param by_user_id: to what user shall snapshot belong? If None, any will match.
        :param by_is_verified: shall the snapshot be moderator-approved? If None, any will match.
        :param by_validation_statuses: searched statuses of snapshot validation. If None, any will match.
        :param by_is_from_robot: shall the snapshot be generated from robot code run (True) or uploaded by human (False)? If None, any will match.
        """
        # at this point locals() will have dict of supplied args and their values
        # using __class__ so we can reuse self=smth from **locals()
        meta = self.__class__.get_latest_meta(**locals())
        _dict = meta.model_dump() if hasattr(meta, "model_dump") else meta.dict()
        return _schemas.Snapshot(
            **_dict,
            result=self.__get_result_by_url_or_id(
                is_loadable=meta.is_result_json_loadable,
                result_file_url=meta.result_file_url,
                snapshot_id=meta.id,
            ),
        )



from datetime import datetime as _datetime
import typing as _typing
from typing import Optional as _Optional
from enum import Enum

from pydantic import BaseModel as __BaseModel, Field as _Field


class SnapshotMeta(__BaseModel):
    user_id: int
    challenge_id: int
    is_verified: bool = False
    verified_message: _Optional[str] = None
    date_created: _datetime
    internal_error: _Optional[str] = (
        None  # might be removed in future. moderator attribute
    )
    id: int
    is_from_code_run: bool
    challenge_name: str
    is_result_json_loadable: _Optional[bool] = _Field(
        default=None, alias="is_json_data_loadable"
    )
    # backend responses with alias name, we rename it into schema name
    result_file_url: _Optional[str] = _Field(default=None, alias="json_data_file_url")


class Snapshot(SnapshotMeta):
    result: _typing.Any = None
    "loaded result of snapshot"


class DataType(str, Enum):
    PRICE = "price"
    AREA = "area"
    FLOOR = "floor"
    ROOMS = "rooms"
    status = "status"
    ID = "id"
    WWW = "www"
    FINISH = "finish"
    CURRENCY = "currency"
    ORIENTATION = "orientation"
    DATETIME = "datetime"


class _ParamDataType(str, Enum):
    pass


class ParamDatetime(_ParamDataType):
    LANGUAGE = "languages"
    'to do not use default one, use "none" or "auto"'


class ParamOrientation(_ParamDataType):
    LOCALE = "locale"




import unittest   # The test framework
from src.dphelper.helper import DPHelper

class TestDPHelperProxy(unittest.TestCase):
   
    def test_default_api(self):
        dphelper = DPHelper(
            is_verbose=True,
            proxy_provider=None,
        )
        www = 'https://dphelper.dataplatform.lt'
        html = dphelper.from_url(www)
        self.assertEqual(html, '{"Hello_from":"backend"}')

if __name__ == '__main__':
    unittest.main()




import unittest   # The test framework
from src.dphelper.helper import DPHelper

class TestDPHelperParsers(unittest.TestCase):

    def test_parse_real_estate_listing(self):
        dphelper = DPHelper()
        self.assertEqual(dphelper.parse_rows(
            ['id', 'floor', 'area', 'rooms', 'price'],
            [['A1', '2', '50.44', '2', '100.000']]
        ), [{'area': 50.44, 'floor': 2, 'id': 'a1', 'price': 100000.0, 'rooms': 2}])

    def test_parse_id(self):
        dphelper = DPHelper()
        self.assertEqual(dphelper.parse_rows(['id'], [['A1']]), [{'id': 'a1'}])
        self.assertEqual(dphelper.parse_rows(['id'], [['a 1']]), [{'id': 'a1'}])

    def test_parse_price(self):
        dphelper = DPHelper()
        self.assertEqual(dphelper.parse_rows(['price'], [['100.000,00']]), [{'price': 100000}])
        self.assertEqual(dphelper.parse_rows(['price'], [['40.23']]), [{'price': 40.23}])

    def test_parse_value(self):
        dphelper = DPHelper()
        self.assertEqual(dphelper.parse_value('100.000,00', 'price'), 100000)

    def test_tansform_map(self):
        dphelper = DPHelper()     
        schema_map = {"kaina": "price", "plotas": "area"}
        data = [{"kaina": 1, "plotas": 2}]
        self.assertEqual(dphelper.transform_map(schema_map, data), [{"price": 1, "area": 2}])

if __name__ == '__main__':
    unittest.main()



from pprint import pprint as _pprint
import typing as _typing
import urllib.parse as _parse
from typing import Optional as _Optional, List as _List
import urllib.parse as _urllib_parse

import requests as _requests

from .connection import FetchModule as __FetchModule
from .snapshot import SnapshotModule as _SnapshotModule
from . import schemas as _schemas
from .schemas import DataType, ParamDatetime, _ParamDataType

_PATH_PARSE = "parse"


class DPHelper(__FetchModule):  # also includes Config
    def __init__(
        self,
        *,
        is_verbose=False,
        proxy_provider=None,
        proxy_params={},
        request_drop_on_failure_count=4,
        request_initial_sleep_after_failure_in_s=3,
        request_sleep_increase_power=2,
        backend_url="https://dphelper.dataplatform.lt",
        dp_url="https://api.dataplatform.lt",
        api_key=None,
        geo_provider=None,
        retry_on_status_codes=[429, "5xx"],
        is_retry_on_blocked=False,
        is_crash_on_blocked=False,
        blocked_treshold=50,
        secrets_file_path="./secrets.dphelper.json",
    ) -> None:
        self._init_config_values()
        self.IS_VERBOSE = bool(is_verbose)
        self.REQUEST_DROP_ON_FAILURE_COUNT = int(request_drop_on_failure_count)
        self.REQUEST_INITIAL_SLEEP_AFTER_FAILURE_IN_S = int(
            request_initial_sleep_after_failure_in_s
        )
        self.REQUEST_SLEEP_INCREASE_POWER = int(request_sleep_increase_power)
        self.set_backend_url(str(backend_url))
        self.set_dp_url(str(dp_url))
        self.snapshot = _SnapshotModule(config=self, fetcher=self, api_key=api_key)
        self.PROXY_PROVIDER = proxy_provider
        self.PROXY_PARAMS = proxy_params
        self.API_KEY = api_key
        self.API_SERVICES_CONFIG = {}
        self.GEO_PROVIDER = geo_provider
        self.RETRYABLE_STATUS_CODE_RANGES = self._make_status_code_ranges(
            retry_on_status_codes
        )
        self.IS_RETRY_ON_BLOCKED = is_retry_on_blocked
        self.IS_CRASH_ON_BLOCKED = is_crash_on_blocked
        self.BLOCKED_TRESHOLD = blocked_treshold
        self.SECRETS_FILE_PATH = secrets_file_path
        self.populate_secrets_from_file()

        if api_key is not None:
            self.authentificate_with_api_key(api_key)

    # conservative - never throw
    def authentificate_with_api_key(self, api_key: str):
        try:
            AUTH_URL = self.DP_URL + "/security/api_services"
            response = _requests.get(AUTH_URL, params={"api_key": api_key})
            if response.status_code == 200:
                self.API_SERVICES_CONFIG = response.json()
        except:
            pass

    def upload_image_from_url(self, url: str, id: _Optional[str] = None) -> str:
        """Uploads image from url to image service. Returns image id"""
        img_service_key = self.get_key_for_image_service()
        if not img_service_key:
            raise ValueError(
                "No image service key found. Please authentificate with API key."
            )

        img_service_params = {
            "url": url,
        }
        if id is not None:
            img_service_params["id"] = id

        encoded_params = _urllib_parse.urlencode(img_service_params)
        img_service_url = "https://images.dataplatform.lt/upload/url?" + encoded_params
        headers = {
            "x-api-key": img_service_key,
        }

        response = _requests.request("POST", img_service_url, headers=headers)
        result = response.json()
        return result

    def get_key_for_image_service(self) -> str:
        return self.API_SERVICES_CONFIG.get("image", "")

    def get_backend_greeting(self) -> str:
        response = _requests.get(self.BACKEND_URL)
        return response.json().get("Hello_from")

    def get_backend_url(self) -> str:
        return self.BACKEND_URL

    def get_dp_url(self) -> str:
        return self.DP_URL

    def __get_utils_url(self, path: str) -> str:
        "joins root url with specified path"
        # appending "/" to root so path gets joined, not overwriten
        # >>> "http://q.w/r" == _parse.urljoin("http://q.w/e", "r")
        # >>> "http://q.w/e/r" == _parse.urljoin("http://q.w/e" + "///////", "r")
        url = _parse.urljoin(self.BACKEND_URL + "/", self.BACKEND_UTILS_PATH)
        return _parse.urljoin(url + "/", path)

    def parse_rows(
        self,
        schema: _List[str],
        data: _List[_List[str]],
        verbose: _Optional[bool] = None,
    ) -> _List[dict]:
        """parses from (look below) to list of dict (look even more below)
        >>> schema = [header1, header2, ...]
        >>> data = [
            [cell1_value, cell2_value, ...],  # 1st row
            [Cell1_value, Cell2_value, ...],  # 2nd row
            ...
        ]

        returns list of dict:
        >>> return_value = [
            {header1: cell1_value, header2: cell2_value}, # 1st row
            {header2: Cell1_value, header2: Cell2_value}, # 2nd row
            ...
        ]

        :param schema: list of headers (eg price, area, floor, id, balcony, status, ...)
        :param data: two-dimensional list of data; It will be parsed as `data[row][column]`. Value has to be json-serializable (string, int, float, not instance of class, ...)
        :param verbose: shall error text be printed?"""
        # local verbose > config verbose
        verbose = self.IS_VERBOSE if verbose is None else verbose

        _data = {
            "schema": schema,
            "data": data,
        }

        response = self._request_or_repeat(
            _requests.post,
            url=self.__get_utils_url(_PATH_PARSE),
            json=_data,
            params={"cell_cnt": len(schema) * len(data)},
        )

        try:
            json_data: dict = response.json()
            is_success = json_data["is_success"]
            if not is_success and verbose:
                _pprint(json_data.get("error"))
            return json_data.get("results", [])
        except Exception:
            raise RuntimeError(
                "Could not understand remote server response for rows-parsing. "
                f"Try updating {self.PACKAGE_NAME} and try again"
            )

    def transform_map(
        self,
        schema_map: dict,
        data: _List[dict],
        parse=True,
        verbose: _Optional[bool] = None,
        *,
        exclude_extra_fields=False,
        parse_also_extra_fields=False,
    ) -> _List[dict]:
        """Renames keys of data, parses on demand.
        >>> schema_map = {"kaina": "price", "plotas": "area"}
        >>> data = [{"kaina": 1, "plotas": 2}]
        >>> return_data = [{"price": 1, "area": 2}] # if parse is False

        :param schema_map: dict where key is old-key, value is new-key (the one which is wanted)
        :param data: list of dicts, whose keys will be renamed, values parsed (if parse=True)
        :param parse: rename schema and parse values? Elsewise, only schema will be renamed
        :param verbose: verbose parse_rows errors?
        :param exlude_extra_fields: should fields, that are not inside schema_map, be excluded from result?
        :param parse_also_extra_fields: should extra fields be parsed? else they will sustain original formatting.
        :return: list of dicts with renamed keys
        """

        keys_old = schema_map.keys()
        keys_extra = set(data[0].keys()) - set(keys_old)
        keys_combined = keys_extra.union(set(keys_old))

        if not exclude_extra_fields:
            # if extra fields not to be excluded, make sure the data is not malformed:
            # each row has exactly same keys

            for i, row in enumerate(data):
                row_keys = set(row.keys())
                if row_keys != keys_combined:
                    extra = row_keys - keys_combined
                    extra_str = f"data[{i}] has extra keys: {extra}; " if extra else ""
                    missing = keys_combined - row_keys
                    missing_str = (
                        f"data[{i}] is missing these keys: {missing}; "
                        if missing
                        else ""
                    )
                    raise RuntimeError(
                        "Keys mismatch in transformation: " f"{extra_str}{missing_str}"
                    )

        if parse:
            is_extra_excluded = exclude_extra_fields or not parse_also_extra_fields
            _keys = keys_old if is_extra_excluded else keys_combined
            unpack_dict = lambda row: [row.get(key) for key in _keys]
            schema = [schema_map[key] if key in keys_old else key for key in _keys]
            parsed_data = self.parse_rows(
                # values of schema_map - to what we want to rename
                schema=schema,
                # for parsing, data has to be without labels, unpacked into array
                data=[unpack_dict(row) for row in data],
                verbose=verbose,
            )

            # parsed data loses unparsed fields. Append them
            if not exclude_extra_fields and not parse_also_extra_fields:
                for new_row, old_row in zip(parsed_data, data):
                    new_row.update({key: old_row.get(key) for key in keys_extra})

            return parsed_data

        # else we will do renaming locally
        return [
            {
                schema_map.get(key) if key in keys_old else key: row.get(key)
                for key in keys_combined
            }
            for row in data
        ]

    @staticmethod
    def update_table(
        data: _List[dict],
        f_check: _typing.Callable[[dict], bool],
        f_update: _typing.Callable[[dict], dict],
    ) -> _List[dict]:
        """updates matching table rows using provided `f_update`. Returns altered table.

        :param data: list of dictionaries - list of rows.
        :param f_check: function that will return `True` if the row has to be altered.
        This function will be checked upon each row.
        :param f_update: function whose returned value will be applied to checked rows.
        :return: Altered version of original table
        """
        results: list[dict] = []
        for row in data:
            result = {}
            result.update(row)
            if f_check(row):
                payload = f_update(row)
                result.update(payload)

            results.append(result)

        return results

    def geocode(
        self, location: str, is_reverse: bool, *, geo_provider: _Optional[str] = ...
    ) -> dict:
        """
        also see `.get_coords(...)`, `.get_address(...)`
        :param location: address or lat,lng
        :param is_reverse: reverse geocode coordinates to address? (Expect coordinates in location?)
        :param geo_provider: "google", "rc", "mixed", "any" or your secret provider
        :return:
        {'address': 'smth', 'lat': 55.55, 'lng': 55.55, ...}"""
        geo_provider = self.GEO_PROVIDER if geo_provider is ... else geo_provider
        geo_provider = None if str(geo_provider).lower() == "any" else geo_provider
        params = {"is_reverse": is_reverse, "provider": geo_provider}
        if self.API_KEY is not None:
            params["api_key"] = self.API_KEY
        response = self._request_or_repeat(
            _requests.post,
            self.__get_utils_url("geocode/"),
            params=params,
            json={"location": str(location)},
        )
        self._response_ok_or_raise(response)
        return response.json()

    @staticmethod
    def __coords_to_location(lat: float, lng: float) -> str:
        return f"{lat},{lng}"

    def get_coords(self, address: str, *, geo_provider: _Optional[str] = ...) -> dict:
        """:param geo_provider: "google", "rc", "mixed", "any" or your secret provider
        >>> .get_coords("Didlaukio g. 59, Vilnius")
        {'lat': 54.7316978, 'lng': 25.2619945}

        """
        location = self.geocode(address, is_reverse=False, geo_provider=geo_provider)
        return {"lat": location.get("lat"), "lng": location.get("lng")}

    def get_address(
        self, lat: float, lng: float, *, geo_provider: _Optional[str] = ...
    ) -> _Optional[str]:
        """:param geo_provider: "google", "rc", "mixed", "any" or your secret provider
        >>> .get_address(54.7316978, 25.2619945)
        'Didlaukio g. 59, 08302 Vilnius, Lithuania'
         >>> # would return None if no address found by coordinates
        """
        location = self.geocode(
            self.__coords_to_location(lat, lng),
            is_reverse=True,
            geo_provider=geo_provider,
        )
        return location.get("address")

    def get_address_components(
        self,
        address: _Optional[str] = None,
        lat: _Optional[float] = None,
        lng: _Optional[float] = None,
        raw=True,
        *,
        geo_provider: _Optional[str] = ...,
    ):
        ':param geo_provider: "google", "rc", "mixed", "any" or your secret provider'
        if not raw:
            raise NotImplemented(
                f"Not raw address components are not supported in current {self.PACKAGE_NAME} version."
            )
        is_reverse = False if address else True
        location_str = address if address else self.__coords_to_location(lat, lng)
        location = self.geocode(
            location_str, is_reverse=is_reverse, geo_provider=geo_provider
        )
        return location.get("raw_address_components")

    def parse_value(
        self,
        value: _typing.Any,
        _type: DataType,
        params: _typing.Dict[_ParamDataType, _typing.Any] = None,
        verbose: _Optional[bool] = None,
    ) -> _typing.Any:
        """parses a single value using datatype and appending its parameters:
        if datatype is DataType.DATETIME and params is {ParamsDatetime.LOCALE: "lt"},
        then datatype will be transformed into `datetime?locale=lt` .
        returns parsed value
        """
        if isinstance(params, dict):
            _new_dict = {}
            for _key, _value in params.items():
                # normalize, to avoid "ParamDatetime.LOCALE" as key
                _key = _key.value if isinstance(_key, _ParamDataType) else str(_key)
                _new_dict[_key] = _value
            params = _new_dict

        # todo: find lib that would handle array values correctly (not as str(array))
        _params = _urllib_parse.urlencode(params) if params else None
        _type = _type.value if isinstance(_type, DataType) else str(_type)
        _type = f"{_type}?{_params}" if _params else _type

        results = self.parse_rows(schema=[_type], data=[[value]], verbose=verbose)
        if results is None:
            return None
        elif len(results) == 0:
            return None
        else:
            row = results[0]
            if len(row) != 1:
                return None
            # return first value because only one value was to be parsed
            return next(iter(row.values()))
